/*
	PROforma, PROGS FOnt & Raster MAnager

	PFPictureRecognize
		try to recognize whether the picture can be displayed by the current
		picture driver

	PFPictureRatio
		get the aspect ratio of a picture (if possible)

	PFPictureColourCount
	PFPictureColourGrayGet
	PFPictureColourRGBGet
	PFPictureColourCMYKGet
		figure out how many colours are used in a picture and what the default
		colours for this picture are

	PFPictureDisplay
	PFPictureDisplayGray
	PFPictureDisplayRGB
	PFPictureDisplayCMYK
		display the given picture at the given position using the specified
		picture driver. the first variant uses the default colours, the other
		variants need a list of colours from the caller

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 18 December 1995
	19.06.1996 clipping in PFPictureElement should now be correct
	10.11.1996 - PageOrigin now adjusted to position at which the picture
				 should be displayed
			   - picture display wrapped in PFGstateSave & PFGstateRestore
	24.11.1996 bug fixed cfr pictures with fixed colours
	27.11.1996 adjusting PageOrigin now correct, and made sure that
			   PFMove(gstate,0,0) in the picture will move to the origin of the
			   picture !
*/

#include "gstate.h"
#include "system.h"
#include "path.h"

extern PICTdriver *pictures;	/* currently supported picture drivers */

#define PICTPREPARE catch( PFGstateSave(gstate) ); \
					gstate->Origin.x=-gstate->posdev.x; \
					gstate->Origin.y=-gstate->posdev.y; \
					gstate->todefaultCTM.xd=0; \
					gstate->todefaultCTM.yd=0; \
					currCTMset(gstate)

static Error find(Gstate gstate, int name, PICTdriver **driver)
{
	Error err=ERR_ITNF; 			/* assume NOT FOUND */
	*driver=NULL;

	/* filter out picture types which exist for compatibility */

	switch(name)
	{
	case PF_PICTURE_QL4SCREEN:
		name=PF_NAME("QL mode 4 screen, 512x256");
		break;
	case PF_PICTURE_QL8SCREEN:
		name=PF_NAME("QL mode 8 screen, 256x256");
		break;
	case PF_PICTURE_PIC:
		name=PF_NAME("QPTR Picture");
		break;
	case PF_PICTURE_PAINTER4:
		name=PF_NAME("The PAINTER, mode 4 compressed picture");
		break;
	case PF_PICTURE_PAINTER8:
		name=PF_NAME("The PAINTER, mode 8 compressed picture");
		break;
	}

	/* actually search for the driver */
	if (name<0)
	{
		/* an actual id has been passed */

		int id;
		for (*driver=pictures, id=-1 ;
			 *driver && id!=name ;
			 *driver=(*driver)->next, id--);
	}
	else
	if (name>50)
	{
		/* search for match on driver's name */

		for (*driver=pictures ;
			 *driver && !STRSameCD((*driver)->name, (char *)name) ;
			 *driver=(*driver)->next);
	}
	if (*driver) err=ERR_OK;

	return err;
}

Error PFPictureRecognize(Gstate gstate, int name, char *picture)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	err=driver->Recognize(gstate,picture);

	return err;
}

Error PFPictureRatio(Gstate gstate, int name, char *picture,
						int *xratio, int *yratio)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	err=driver->AspectRatio(gstate,picture,xratio,yratio);

	return err;
}

Error PFPictureColourCount(Gstate gstate, int name, char *picture, int *count)
{
	Error err;
	PICTdriver *driver;
	int space;

	catch( find(gstate,name,&driver) );
	err=driver->ColourInfo(gstate,picture,count,&space);

	return err;
}

/* --- Get the default colours which are used in a picture --- */

typedef struct {
	Error (*Gray)(pt *from, void *to);
	Error (*RGB)(ColourRGB *from, void *to);
	Error (*CMYK)(ColourCMYK *from, void *to);
} ColConv;

Error Gray2Gray(pt *from, void *to);
Error Gray2RGB(pt *from, void *to);
Error Gray2CMYK(pt *from, void *to);
Error RGB2Gray(ColourRGB *from, void *to);
Error RGB2RGB(ColourRGB *from, void *to);
Error RGB2CMYK(ColourRGB *from, void *to);
Error CMYK2Gray(ColourCMYK *from, void *to);
Error CMYK2RGB(ColourCMYK *from, void *to);
Error CMYK2CMYK(ColourCMYK *from, void *to);

static ColConv convtoGray = { Gray2Gray, RGB2Gray, CMYK2Gray };
static ColConv convtoRGB  = { Gray2RGB , RGB2RGB , CMYK2RGB  };
static ColConv convtoCMYK = { Gray2CMYK, RGB2CMYK, CMYK2CMYK };

static Error getcolours(Gstate gstate, PICTdriver *driver, char *picture,
						void *colours, int size, ColConv *conv)
{
	int count, space, i;
	Error err;

	catch( driver->ColourInfo(gstate,picture,&count,&space) );
	if (count==0) return ERR_ISYN;

	switch (space)
	{
	case PF_COLOURSPACE_GRAYSHADE:
		for (i=0 ; i<count ; i++)
		{
			pt colour;
			catch( driver->ColourGet(gstate,picture,i,&colour) );
			conv->Gray(&colour,colours);
			((char *)colours)+=size;
		}
		break;
	case PF_COLOURSPACE_RGB:
		for (i=0 ; i<count ; i++)
		{
			ColourRGB colour;
			catch( driver->ColourGet(gstate,picture,i,&colour) );
			catch( conv->RGB(&colour,colours) );
			((char *)colours)+=size;
		}
		break;
	case PF_COLOURSPACE_CMYK:
		for (i=0 ; i<count ; i++)
		{
			ColourCMYK colour;
			catch( driver->ColourGet(gstate,picture,i,&colour) );
			catch( conv->CMYK(&colour,colours) );
			((char *)colours)+=size;
		}
		break;
	default:
		err=ERR_IEXP;
		break;
	}
	return err;
}

Error PFPictureColourGrayGet(Gstate gstate, int name, char *picture,
						pt *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	err=getcolours(gstate, driver, picture, colours, sizeof(pt), &convtoGray);
	return err;
}

Error PFPictureColourRGBGet(Gstate gstate, int name, char *picture,
						ColourRGB *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	err=getcolours(gstate, driver, picture, colours, sizeof(ColourRGB), &convtoRGB);
	return err;
}

Error PFPictureColourCMYKGet(Gstate gstate, int name, char *picture,
						ColourCMYK *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	err=getcolours(gstate, driver, picture, colours, sizeof(ColourCMYK), &convtoCMYK);
	return err;
}

/* --- actually display a picture --- */

static selectGray(Gstate gstate, int which)
{
	if (gstate->pictcolcurr!=which)
	{
		gstate->pictcolcurr=which;
		return PFColourGray(gstate,gstate->pictcoltable.gray[which]);
	}
	else return ERR_OK;
}

Error PFPictureDisplayGray(Gstate gstate, int name, char *picture,
						pt xsiz, pt ysiz, pt *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	PFPathClear(gstate);
	gstate->pictcolcurr=-1;
	gstate->pictcoltable.gray=colours;

	PICTPREPARE;
	err=driver->Display(gstate,picture,xsiz,ysiz,selectGray);
	PFGstateRestore(gstate);
	return err;
}

static selectRGB(Gstate gstate, int which)
{
	if (gstate->pictcolcurr!=which)
	{
		gstate->pictcolcurr=which;
		return PFColourRGB(gstate,&gstate->pictcoltable.RGB[which]);
	}
	else return ERR_OK;
}

Error PFPictureDisplayRGB(Gstate gstate, int name, char *picture,
						pt xsiz, pt ysiz, ColourRGB *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	PFPathClear(gstate);
	gstate->pictcolcurr=-1;
	gstate->pictcoltable.RGB=colours;

	PICTPREPARE;
	err=driver->Display(gstate,picture,xsiz,ysiz,selectRGB);
	PFGstateRestore(gstate);
	return err;
}

static selectCMYK(Gstate gstate, int which)
{
	if (gstate->pictcolcurr!=which)
	{
		gstate->pictcolcurr=which;
		return PFColourCMYK(gstate,&gstate->pictcoltable.CMYK[which]);
	}
	else return ERR_OK;
}

Error PFPictureDisplayCMYK(Gstate gstate, int name, char *picture,
						pt xsiz, pt ysiz, ColourCMYK *colours)
{
	Error err;
	PICTdriver *driver;

	catch( find(gstate,name,&driver) );
	PFPathClear(gstate);
	gstate->pictcolcurr=-1;
	gstate->pictcoltable.CMYK=colours;

	PICTPREPARE;
	err=driver->Display(gstate,picture,xsiz,ysiz,selectCMYK);
	PFGstateRestore(gstate);
	return err;
}

Error PFPictureDisplay(Gstate gstate, int name, char *picture, pt xsiz, pt ysiz)
{
	Error err;
	PICTdriver *driver;
	void *colours=NULL;
	int count,space;

	catch( find(gstate,name,&driver) );
	catch( driver->ColourInfo(gstate,picture,&count,&space) );
	PICTPREPARE;

	if (count==0)
		err=driver->Display(gstate,picture,xsiz,ysiz,(Error (*)(Gstate,int))NULL);
	else switch (space)
	{
	case PF_COLOURSPACE_GRAYSHADE:
		catch( MEMAllocate(count*sizeof(pt),&colours) );
		err=PFPictureColourGrayGet(gstate,name,picture,(pt *)colours);
		if (err) break;
		gstate->pictcolcurr=-1;
		gstate->pictcoltable.gray=(pt *)colours;
		err=driver->Display(gstate,picture,xsiz,ysiz,selectGray);
		break;
	case PF_COLOURSPACE_RGB:
		catch( MEMAllocate(count*sizeof(ColourRGB),&colours) );
		err=PFPictureColourRGBGet(gstate,name,picture,(ColourRGB *)colours);
		if (err) break;
		gstate->pictcolcurr=-1;
		gstate->pictcoltable.RGB=(ColourRGB *)colours;
		err=driver->Display(gstate,picture,xsiz,ysiz,selectRGB);
		break;
	case PF_COLOURSPACE_CMYK:
		catch( MEMAllocate(count*sizeof(ColourCMYK),&colours) );
		err=PFPictureColourCMYKGet(gstate,name,picture,(ColourCMYK *)colours);
		if (err) break;
		gstate->pictcolcurr=-1;
		gstate->pictcoltable.CMYK=(ColourCMYK *)colours;
		err=driver->Display(gstate,picture,xsiz,ysiz,selectCMYK);
		break;
	}
	MEMRelease(colours);
	PFGstateRestore(gstate);
	return err;
}

/* --- support routines for drawing pictures --- */

Error PFPictureElement(Gstate gstate, pt xsiz, pt ysiz, pt xorg, pt yorg)
{
	Bitmap raster=gstate->raster;

	if (gstate->currCTM.xr || gstate->currCTM.yr)
	{
		/* difficult case - rotated and/or slanted rectangle */
		coor curr, hor, ver;
		Error err;

		curr.x=xorg;
		curr.y=yorg;
		user2dev(gstate,&curr);
		hor.x=xsiz;
		ver.y=ysiz;
		hor.y=ver.x=0;
		reluser2dev(gstate,&hor);
		reluser2dev(gstate,&ver);

		addflatline(gstate, curr.x			  , curr.y			  , curr.x+hor.x	  , curr.y+hor.y);
		addflatline(gstate, curr.x+hor.x	  , curr.y+hor.y	  , curr.x+hor.x+ver.x, curr.y+hor.y+ver.y);
		addflatline(gstate, curr.x+hor.x+ver.x, curr.y+hor.y+ver.y, curr.x+ver.x	  , curr.y+ver.y);
		addflatline(gstate, curr.x+ver.x	  , curr.y+ver.y	  , curr.x			  , curr.y);
		err=fillpath(gstate, raster->bitdraw->TrapezoidHorizontal, raster, PF_WINDING_RULE);
		PFPathClear(gstate);
		return err;
	}
	else
	{
		/* simple case - real rectangle */
		coor pos;
		pt xend=xorg+xsiz;
		pt yend=yorg+ysiz;
		pt dummy;

		pos.x=xorg;
		pos.y=yorg;
		user2dev(gstate,&pos);
		xorg=pt2rshort(pos.x);
		yorg=pt2rshort(pos.y);

		pos.x=xend;
		pos.y=yend;
		user2dev(gstate,&pos);
		xend=pt2rshort(pos.x);
		yend=pt2rshort(pos.y);
		if (xend<xorg) { dummy=xorg; xorg=xend; xend=dummy; }
		if (yend<yorg) { dummy=yorg; yorg=yend; yend=dummy; }

		/* check for complete invisibility - allows faster picture drawing */
		if ( xorg==xend || yorg==yend ||
			 xend < 0 || yend < 0 ||
			 xorg > raster->clip.xsiz ||
			 yorg > raster->clip.ysiz )
			return ERR_ORNG;

		return raster->bitdraw->Block(gstate->raster,
							xend-xorg, yend-yorg, xorg, yorg);
	}
}
