/*
	header file for PROforma, PROGS Font & Raster Manager
	definitions of (external) PROforma drivers

	DRIVdriver	drivers for screen ("SCR ") and printer ("PRT ")
				to display the bitmap on the device
	BITMdriver	driver for drawing in bitmaps ("BITM")
				to display something in the bitmap (basic drawing routines)
	PICTdriver	driver for handling pictures ("PICT")
				to display a picture in a gstate

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 24 September 1995
	07.02.1996 added Handle routine for extensions to all drivers

*/

#include "win.h"

#define PF_SCREENDRIVER 	0x53435220	/* "SCR " */
#define PF_PRINTERDRIVER	0x50525420	/* "PRT " */
#define PF_BITMAPDRIVER 	0x4249544d	/* "BITM" */
#define PF_PICTUREDRIVER	0x50494354	/* "PICT" */

#define BITMAP_NO_BYTEGROUPING		0
#define BITMAP_SEPARATE_PLANES		0
#define BITMAP_NO_PLANES			0
#define BITMAP_DIRECTION_HORIZONTAL 12
#define BITMAP_DIRECTION_VERTICAL	3

typedef struct {            /* bitmap are always line oriented */
	char PlaneCount;		/* number of colour planes in bitmap */
	char PlaneDepth;		/* number of bits for each plane */
	char ByteGrouping;		/* number of bytes grouped in ByteDirection */
	char PlaneGrouping; 	/* number of consecutive bytes in colour plane */
} BitmapType;

typedef struct _ANYdriver {
	struct _ANYdriver *next;	/* next driver in list of drivers */
	int identifier; 			/* identifies type of driver */
} ANYdriver;


/* printer or screen driver definition */

#define PF_COLOURSPACE_GRAYSHADE 1	/* mono device - black and white only */
#define PF_COLOURSPACE_RGB		 2	/* additive colour device */
#define PF_COLOURSPACE_CMY		 3	/* subtractive colour support - no black */
#define PF_COLOURSPACE_CMYK 	 4	/* subtractive colour support with black */

typedef struct {
	Error (*ColourAdjust)(Gstate, ColourRGB *);
	Error (*ColourNearest)(Gstate, ColourRGB *, int *);
	Error (*ColourDelta)(Gstate, ColourRGB *, ColourRGB *);
} RGBRouts;
typedef struct {
	Error (*ColourAdjust)(Gstate, ColourCMYK *);
	Error (*ColourNearest)(Gstate, ColourCMYK *, int *);
	Error (*ColourDelta)(Gstate, ColourCMYK *, ColourCMYK *);
} CMYKRouts;
typedef struct {
	Error (*ColourAdjust)(Gstate, pt *);
	Error (*ColourNearest)(Gstate, pt *, int *);
	Error (*ColourDelta)(Gstate, pt *, pt *);
} GrayRouts;

typedef struct _DRIVdriver {
	struct _DRIVdriver *next;	/* next driver in list of drivers */
	int identifier; 			/* identifies type of driver */

	char name[PF_MAXDRIVERNAME];/* name of driver  - the name of a driver */
								/* should not start with "-" or " " !! */
	int xdpi, ydpi; 			/* device resolution, dots per inch */
	BitmapType bitmap;			/* type of bitmap which device needs */

	/* specific routines for handling the driver */
	Error (*InitDriver)();
	Error (*Define)(char *name, char *value);

	/* routines to complete - finalise gstate */
	Error (*InitInit)(Gstate, char *device, PBbox *area);
	Error (*BufferSize)(Gstate, int xsiz, int ysiz, Size *size);
	Error (*InitDone)(Gstate);
	Error (*CleanUp)(Gstate);

	/* routines to display the page */
	Error (*InitPage)(Gstate);
	Error (*ShowPage)(Gstate);
	Error (*StopPage)(Gstate);

	/* routines to control the displaying of pages */
	Error (*SetCopies)(Gstate, int count);
	Error (*SetDisplayMode)(Gstate, int mode);
	Error (*WindowMove)(Gstate, int dx, int dy);

	Error (*BufferUpdate)(Gstate);
	Error (*BufferSave)(Gstate);
	Error (*BufferRestore)(Gstate);

	/* colour handling */
	short ColourSpace;			/* colourspace which is used */
	union {                     /* routines to control the colour */
		RGBRouts RGB;			/* which is used for this device */
		CMYKRouts CMYK;
		GrayRouts Gray;
	} colour;

	/* driver info */
	Error (*GetFullPageBbox)(Gstate, Bbox *box);
	Error (*DefaultDevice)(Gstate, char *device);

	/* future use */
	Error (*Handle)(int command, ...);
} DRIVdriver;


/* bitmap driver definition */

typedef struct {
	/* drawing routines */
	Error (*Pixel)(Bitmap, short x, short y);
	Error (*LineHorizontal)(Bitmap, short aty, short fromx, short tox);
	Error (*LineVertical)(Bitmap, short atx, short fromy, short toy);
	Error (*Block)(Bitmap, short xsiz, short ysiz, short xorg, short yorg);
	Error (*TrapezoidHorizontal)(Bitmap, short starty, short endy,
					pt leftstartx, pt leftincr, pt rightstartx, pt rightincr);
	Error (*TrapezoidVertical)(Bitmap, short startx, short endx,
					pt topstarty, pt topincr, pt bottomstarty, pt bottomincr);
	Error (*HairLine)(Bitmap, pt fromx, pt fromy, pt tox, pt toy);
	Error (*DrawMask)(Bitmap, Mask *mask, short x, short y);
	Error (*ChangeOrigin)(Bitmap, short x, short y);
} BitDrawRouts;

typedef struct _BITMdriver {
	struct _BITMdriver *next;	/* next driver in list of drivers */
	int identifier; 			/* identifies type of driver */

	BitmapType type;			/* type of bitmap defined here */

	/* Sets of drawing routines for bitmap */
	/* A FastPattern is a 8x8 pattern (or multiple) */
	BitDrawRouts Black;
	BitDrawRouts White;
	BitDrawRouts FastPattern;
	BitDrawRouts Pattern;

	/* set/read pixel cursor */
	Error (*CursorSet)(Bitmap, short x, short y);
	Error (*CursorMove)(Bitmap, short x, short y);
	Error (*CursorGet)(Bitmap, short *x, short *y);

	/* direct pixel colour set/read commands */
	Error (*PixelRead)(Bitmap, int *colour);
	Error (*PixelReadNextX)(Bitmap, int *colour);
	Error (*PixelReadPrevX)(Bitmap, int *colour);
	Error (*PixelReadNextY)(Bitmap, int *colour);
	Error (*PixelReadPrevY)(Bitmap, int *colour);
	Error (*PixelSet)(Bitmap, int colour);
	Error (*PixelSetNextX)(Bitmap, int colour);
	Error (*PixelSetPrevX)(Bitmap, int colour);
	Error (*PixelSetNextY)(Bitmap, int colour);
	Error (*PixelSetPrevY)(Bitmap, int colour);

	/* copy (part of) a bitmap to (another) bitmap */
	Error (*CopyArea)(Bitmap from, Bitmap to, short xsiz, short ysiz,
					  short fromx, short fromy, short tox, short toy);

	/* info about the bitmap */
	Error (*BufferSize)(short xsiz, short ysiz, Size *size);
	int FastPatternXsize, FastPatternYsize;
	char FastestDirection;

	/* future extension */
	Error (*Handle)(int command, ...);
} BITMdriver;


/* picture driver definition */

typedef struct _PICTdriver {
	struct _PICTdriver *next;	/* next driver in list of drivers */
	int identifier; 			/* identifies type of driver */

	char name[PF_MAXDRIVERNAME];/* name of driver  - the name of a driver */
								/* should not start with "-" or " " !! */

	/* can we recognize the picture */
	Error (*Recognize)(Gstate, char *base);
	Error (*AspectRatio)(Gstate, char *base, int *xratio, int *yratio);

	/* display the picture */
	Error (*Display)(Gstate, char *base, pt xsiz, pt ysiz,
					 Error (*ColourSelect)(Gstate, int));

	/* info on colours used in the picture */
	Error (*ColourInfo)(Gstate, char *base, int *count, int *colourspace);
	Error (*ColourGet)(Gstate, char *base, int which, void *colour);

	/* future extension */
	Error (*Handle)(int command, ...);
} PICTdriver;
